/*
 * $Id: BufferedRandomFileInputStream.java,v 1.1 1999/08/03 11:31:20 gunterma Exp $
 *
 * Copyright (c) 1999  Klaus Guntermann/TU Darmstadt, FB 20, FG SP
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

package de.tu_darmstadt.sp.util;

import java.io.*;
import java.util.*;

/**
 * A buffering mechanism for input files, which need to be seeked.
 * <p>
 * Java does not come with a class that allows random file accesses on
 * reading while doing buffering. Because reads are native reads by
 * the operating system reading larger files in small chunks gives a
 * very bad performance. One wishes to buffer also reads from random
 * access files, even if this would disallow writing to them
 * simultaneously.
 * <br>
 * Of course there is a tradeoff involved, because if you do many
 * seeks and only small reads, and even do not cover the whole foile,
 * any buffering can result in worse performance, because buffered
 * reads would always read too much.
 *
 * @version $Revision: 1.1 $
 * @author Klaus Guntermann <guntermann@iti.informatik.tu-darmstadt.de>
 */
public class BufferedRandomFileInputStream extends InputStream  {
/*
 * This implementation tries to overcome the problem of taking almost
 * no advantage from larger buffers by introducing a list of buffers.
 * The number of buffers and the size of the buffers is configurable.
 * The defaults are 20 buffers of 8192 bytes each.
 * The buffers are reused in a LRU strategy.
 *
 * If there is a reasonable locality in seeking the file, one can expect
 * a good speedup, because later reads will be serviced from one of the
 * buffers. If you do not achieve speedups, try enlarging the number of
 * buffers, if your application reseeks areas of the file at all.
 *
 * The methods <code>mark</code> and <code>reset</code> are not supported
 * by this implementation, because one can seek arbitrary positions.
 * And access to a recently read area is not combined with a large penalty.
 */
  private long currentReadPosition = 0; // our internal read pointer
  private RandomAccessFile below = null; // the "real" file
  private long fileLength = 0; // size of the file after open
  private int blocksize = 8192; // default size of buffered blocks
  private int maxblocks = 20; // default maximum number of blocks

  private Vector buffervector = null; // our cache of buffers
  // the buffer vector entries are structured as follows:
  private class BufferEntry {
    long startPosition; // where this block belongs
    int length; // its current length, which may be less than buffer.length()
    byte [] buffer; // the cached data
  }
  private BufferEntry current = null; // the most entry currently in use

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file named by the path name
   * <code>name</code> in the file system. A new
   * <code>RandomAccessFile</code> is opened for this file, which will
   * be used for all real file operations. Buffering will use a
   * default number of blocks and a default blocksize.
   * @param name the system-dependent file name.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(String name) throws IOException {
    below = new RandomAccessFile(name, "r");
    fileLength = below.length();
    currentReadPosition = -1;
    buffervector = new Vector();
  }

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file named by the path name
   * <code>name</code> in the file system. A new
   * <code>RandomAccessFile</code> is opened for this file, which will
   * be used for all real file operations. The buffering will work
   * with the indicated number <code>numblocks</code> of buffers and
   * will use blocks of size <code>blocklen</code>.
   * @param name the system-dependent file name.
   * @param numblocks the number of buffer blocks to maintain.
   * @param blocklen the maximum length of each buffer.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(String name, int numblocks,
				       int blocklen) throws IOException {
    this(name);
    blocksize = blocklen;
    maxblocks = numblocks;
  }

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file named by the path name
   * <code>name</code> in the file system. A new
   * <code>RandomAccessFile</code> is opened for this file, which will
   * be used for all real file operations. The buffering will work
   * with the indicated number <code>numblocks</code> of buffers and
   * use a default size for the blocks.
   * @param name the system-dependent file name.
   * @param numblocks the number of buffer blocks to maintain.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(String name, int numblocks)
    throws IOException {
    this(name);
    maxblocks = numblocks;
  }

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file specified by the
   * <code>File</code> argument. A new <code>RandomAccessFile</code>
   * is opened for this file, which will be used for all real file
   * operations. Buffering will use a default number of blocks and a
   * default blocksize.
   * @param file the file object.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(File file)
    throws IOException  {
    this(file.getPath());
  }

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file specified by the
   * <code>File</code> argument. A new <code>RandomAccessFile</code>
   * is opened for this file, which will be used for all real file
   * operations. The buffering will work with the indicated number
   * <code>numblocks</code> of buffers and will a default blocksize.
   * @param name the system-dependent file name.
   * @param numblocks the number of buffer blocks to maintain.
   * @param blocklen the maximum length of each buffer.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(File file, int numblocks)
    throws IOException
  {
    this(file);
    maxblocks = numblocks;
  }

  /**
   * Creates a <code>BufferedRandomFileInputStream</code> by opening a
   * connection to an actual file, the file specified by the
   * <code>File</code> argument. A new <code>RandomAccessFile</code>
   * is opened for this file, which will be used for all real file
   * operations. The buffering will work with the indicated number
   * <code>numblocks</code> of buffers and will use blocks of size
   * <code>blocklen</code>.
   * @param name the system-dependent file name.
   * @param numblocks the number of buffer blocks to maintain.
   * @param blocklen the maximum length of each buffer.
   * @exception IOException if the <code>RandomAccessFile</code>
   * throws any exception.
   */
  public BufferedRandomFileInputStream(File file, int numblocks, int blocklen)
    throws IOException
  {
    this(file);
    maxblocks = numblocks;
    blocksize = blocklen;
  }

  /**
   * Closes the file input stream and releases the accsicated
   * <code>RandomAccessFile</code>.
   * @exception IOException is raised, when the underlying file has a
   * problem.
   */
  public void close() throws IOException {
      below.close();
      below = null;
      buffervector = null;
  }

  /**
   * Sets the file-pointer offset, measured from the beginning of this
   * file, at which the next read or write occurs. The file pointer
   * must not be negative or beyond the size of the file. Seeking will
   * ensure, that an appropriate slice of the file is read into a
   * buffer. The read pointer will be positioned accordingly. Reading
   * the next byte after a <code>seek</code> will not require an I/O
   * operation. <code>seek</code> will not do an I/O operation, if the
   * requested position is still avialable in one of the buffers.
   * @param pos the offset position to which to set the read pointer.
   * @exception IOException any problems the underlying
   * <code>RandomAccessFile</code> chooses to report or positioning to
   * a negative position.
   */
  public void seek(long pos) throws IOException {
    long startPosition = (pos/blocksize)*blocksize; // map to start of buffer
    boolean notfound = true; // search control
    int readlength; // how much do we want to read now

    if (pos < 0) {
      throw new IOException("invalid position " + pos + " for seek");
    }
    // find the buffer, which contains the destination position, if any.
    for (int i = 0; i < buffervector.size(); i++) {
      if (((BufferEntry)buffervector.elementAt(i)).startPosition ==
	    startPosition) {
	// move this element to the first position of our
	// buffervector.
	current = (BufferEntry) buffervector.elementAt(i);
	buffervector.removeElementAt(i);
	buffervector.insertElementAt(current,0);
	notfound = false;
	break;
      }
    }
    // if we did not find one, make up a new buffer
    if (notfound) {
      // if we have already too many buffers, remove the last one, but
      // reuse the already allocated memory
      // Because we implement a "move to front" policy, this
      // will be the one which was not used for a long time now.
      if (buffervector.size() >= maxblocks) {
	current = (BufferEntry) buffervector.elementAt(buffervector.size() - 1);
	buffervector.removeElementAt(buffervector.size() - 1);
      } else {// we should get another new buffer
	current = new BufferEntry();
	current.buffer = new byte[blocksize];
      }
      // now we can add a new buffer at front...
      current.startPosition = startPosition;
      below.seek(startPosition);
      readlength = blocksize;
      if (startPosition + readlength > fileLength) {
	readlength = (int) (fileLength - startPosition);
      }
      current.length = readlength;
      below.readFully(current.buffer, 0, readlength);
      buffervector.insertElementAt(current, 0);
    }
    // Now we have the data in the buffer, so just set the read
    // pointer to the destination position.
    currentReadPosition = pos;
  }

  /**
   * Returns the number of bytes that can be read without blocking
   * @return number of bytes available.
  */
  public int available() {
    if (currentReadPosition == -1) return 0; // do not know anything now
    if (currentReadPosition >= fileLength) return 0; // already at EOF
    // say how much we have available in the current buffer
    return (int) (current.length - (currentReadPosition
                                    - current.startPosition));
  }


  /**
   * Returns the length of the file. Because the length cannot change
   * after the file was opened, this request will be serviced from
   * internal information and cannot lead to an exception.
   */
  public long length() {
    return fileLength;
  }

  /**
   * Read a byte from the file. The byte is returned as an integer in
   * the range 0 to 255 (<code>0x00-0xff</code>). This conforms to the
   * specification of <code>RandomAccessFile.read()</code>.
   */
  public int read() throws IOException {
    int data;

    if (currentReadPosition == -1) seek(0); // must initialize...
    if (currentReadPosition >= fileLength) return -1; // EOF
    // if we ran out of the buffer, seek to the proper position.
    if (currentReadPosition >= current.startPosition + current.length) {
      seek(currentReadPosition);
    }
    // pick the data from the current position
    data = current.buffer[(int) (currentReadPosition - current.startPosition)];
    // increment the read pointer
    currentReadPosition += 1;
    return (data & 0xff);
  }

  /**
   * Read up to len bytes from the file into an array of bytes.
   * @param b the buffer to read into.
   * @param off the offset to start storing bytes.
   * @param len the maximum number of bytes to read.
   * @return the number of bytes read or -1, if we are already at the end
   * of the file.
   * @exception IOException any I/O error from the real file.
   */
  public int read(byte [] b, int off, int len) throws IOException {
    int returncode = -1;
    int toread;

    if (currentReadPosition == -1) seek(0); // must initialize...
    if (currentReadPosition >= fileLength) return -1; // EOF
    // if we ran out of the buffer, seek to the proper position.
    if (currentReadPosition >= current.startPosition + current.length) {
	seek(currentReadPosition);
    }
    // pick the data from the current position. Return the remaining
    // bytes from this buffer or the complete length requested.
    toread = (int) (current.length - (currentReadPosition
             - current.startPosition));
    // do not read more than requested
    if (len < toread) toread = len;
    // do not read more than the buffer can take
    if (off + b.length < toread) toread = b.length - off;
    System.arraycopy(current.buffer,
		     (int) (currentReadPosition - current.startPosition),
		     b, off, toread);
    // increment the read pointer
    currentReadPosition += toread;
    return toread;
  }

  /**
   * Attempts to skip the given number of bytes in the file,
   * discarding the skipped bytes. This method may skip over some
   * smaller number of bytes, possibly zero. This may result from any
   * of a number of conditions; reaching end of file before n bytes
   * have been skipped is only one possibility. This method never
   * throws an <code>EOFException</code>. The actual number of bytes skipped is
   * returned. If n is negative, no bytes are skipped.
   * @param n the number of bytes to skip.
   * @exception IOException if an I/O error is reported.
   */
  public long skip(long n) throws IOException {
    if (n < 0) return 0; // should not move back, according to spec's
    if (currentReadPosition == -1) seek(0); // must initialize...
    currentReadPosition += n;
    // check for end of file position
    if (currentReadPosition > fileLength) {
      n -= currentReadPosition - fileLength;
      currentReadPosition = fileLength;
    }
    return n;
  }

  /**
   * Returns the current offset in the file.
   * @return the current read position in the file in bytes from the
   * beginning of the file.
   */
  public long getFilePointer() {
    if (currentReadPosition == -1) return 0; // not yet initialized properly
    return currentReadPosition;
  }


  /**
   * The main routine will just provide a small test. After reading
   * random parts from a file named in the argv[0] we will make a
   * copy of it with the name <code>copy</code>.
   */
  public static void main(String argv[]) {
    BufferedRandomFileInputStream raf;
    BufferedOutputStream of;
    Random r = new Random(); // try at random positions
    long fileLength; // needed for destination calculation
    long pos; // the position to read at
    byte [] buf; // a buffer for blocked reads
    int n; // the number of bytes read
    int count; // for some short statistics
    String filename="scratchfile.for.test";
    try {
      System.out.println("Creating testfile " + filename + ".");
      of=new BufferedOutputStream(new FileOutputStream(filename));
      for (int i = 0; i < 50*8192; i++) {
	of.write(i & 0xff);
      }
      of.close();
      System.out.println("Starting random access.");
      raf = new BufferedRandomFileInputStream(filename);
      fileLength = raf.length();
      System.out.println("Opened file " + filename + ", size is " + fileLength);
      count = 0;
      for (int i = 0 ; i < 200; i++) {
	pos = (long) (r.nextFloat() * fileLength);
	raf.seek(pos);
	n=raf.read();
	if ((pos & 0xff) != n) {
	  System.out.println("Bad match at " + pos + ", got " + n + "/"
              + (pos & 0xff));
	} else count += 1;
      }
      System.out.println("Seeking worked at " + count + " positions.");
      buf = new byte[1234];
      count = 0;
      raf.seek(0);
      of=new BufferedOutputStream(new FileOutputStream(filename + ".copy"));
      while ((n = raf.read(buf)) > 0) {
	of.write(buf, 0, n);
	count += 1;
      }
      of.close();
      System.out.println("Copied file in " + count + " chunks ("
         + fileLength / 1234 + ")");
    } catch (IOException e) {
      System.out.println("Oops, exception " + e);
    }
  }
}
