// $Id: BufferedRandomFileInputStreamTest.java,v 1.1 1999/08/03 11:31:22 gunterma Exp $
// =====================================================================
//
// (history at end)
//

package de.tu_darmstadt.sp.util;

// used packages
import java.lang.*;
import java.util.*;
import java.io.*;
import junit.framework.*;

/**
 * JUnit test cases for class BufferedRandomFileInputStream.
 *
 * @version	$Revision: 1.1 $
 * @author	Klaus Guntermann
 */
public
class BufferedRandomFileInputStreamTest extends TestCase {

  // fixture
  String filename="workfile.for.JUnit.test";
  BufferedRandomFileInputStream brf;

  /**
   * Construct test with a given name.
   * @param name test id
   */
  public BufferedRandomFileInputStreamTest( String name )
  {
    super(name);
  }

  /**
   * Setup fixture.
   */
  protected
  void setUp() {
    // Create a test file to manipulate and open it.
    File infile = null;
    BufferedOutputStream outfile = null;
    boolean ok = false;


    infile = new File(filename);
    if (infile.canRead()) {
      fail("File " + filename + " exists; remove it and try again.");
    } else {
      try {
	outfile = new BufferedOutputStream(new FileOutputStream(filename));
	// Write a pattern into the file to be able to see, whether we
	// are at the correct relative position. This does not help if
	// we have an offset error of a multiple of 256 though.
	for (int i = 0; i < 50*8192; i++) {
	  outfile.write(i & 0xff);
	}
	outfile.close();
	brf = new BufferedRandomFileInputStream(filename);
      } catch (IOException e) {
	fail("Unexpected exception " + e);
      }
    }
  }

  /**
   * Teardown fixture.
   */
  protected
  void tearDown() {
    // Close and remove the test file.
    File infile = null;
    try {
      brf.close();
    } catch (IOException e) {
      fail("Unexpected exception " + e);
    }
    infile = new File(filename);
    if (!infile.delete()) {
      fail("Sorry, cannot delete workfile " + filename);
    }
  }

  /**
   * Simple test of seeks.
   */
  public
  void testSeek() {
    long size = brf.length();
    long half = size>>1;

    // position to start, middle and end of file and check, where we are.
    try {
      brf.seek(0);
      assertEquals("seek to start", (long)0, brf.getFilePointer());
      brf.seek(half);
      assertEquals("seek to middle", half, brf.getFilePointer());
      brf.seek(size);
      assertEquals("seek to end", size, brf.getFilePointer());
    } catch (IOException e) {
      System.out.println(e);
    }
  }

  /**
   * Simple tests of available.
   */
  public
  void testAvailable() {
    long size = brf.length();
    long half = size>>1;
    int amount;

    // position to start, middle and end of file and check, whether
    // there is data available.
    try {
      brf.seek(0);
      amount = brf.available();
      if (amount <= 0) fail("nothing available at start of file");
      brf.seek(half);
      amount = brf.available();
      if (amount <= 0) fail("nothing available at middle of file");
      brf.seek(size);
      assertEquals("available at EOF", (int)0, brf.available());
    } catch (IOException e) {
      System.out.println(e);
    }
  }

  /**
   * Simple tests of random read
   */
  public
  void testRandomRead() {
    // seek to random positions and check single bytes.
    Random r = new Random(); // try at random positions
    long maxseek = brf.length()-1; // do not try to read at EOF
    long pos;
    int n;

    try {
      for (int i = 0 ; i < 200; i++) {
	pos = (long) (r.nextFloat() * maxseek);
	brf.seek(pos);
	n=brf.read();
	if ((pos & 0xff) != n) {
	  fail("Bad random read match at " + pos + ", got " + n + "/"
	       + (pos & 0xff));
	}
      }
    } catch (IOException e) {
      System.out.println(e);
    }
  }

  /**
   * Simple test of sequential block read into arbitrary internal buffer.
   */
  public
  void testCompleteBlockRead() {
    byte [] buf = new byte[1234];;
    long pos;
    int n;

    try {
      brf.seek(0);
      pos = 0;
      while ((n = brf.read(buf)) >= 0) {
	for (int i = 0; i < n; i++) {
	  if ((pos & 0xff) != (buf[i] & 0xff)) {
	    fail("Bad sequential read match at " + pos + ", got "
                 + (buf[i] & 0xff) + "/" + (pos & 0xff));
	  }
	  pos += 1;
	}
      }
      assertEquals("pos should be at end of file now", pos, brf.length());
      assertEquals("FP should be at end of file now", pos,
		   brf.getFilePointer());
    } catch (IOException e) {
      fail("Unexpected exception " + e);
    }
  }

  /**
   * Simple test of sequential byte read.
   */
  public
  void testCompleteByteRead() {
    int b;
    long pos;

    try {
      brf.seek(0);
      pos = 0;
      while ((b = brf.read()) >= 0) {
	if ((pos & 0xff) != b) {
	  fail("Bad sequential read match at " + pos + ", got "
+ b + "/" + (pos & 0xff));
	}
	pos += 1;
      }
      assertEquals("pos should be at end of file now", pos, brf.length());
      assertEquals("FP should be at end of file now", pos,
		   brf.getFilePointer());
    } catch (IOException e) {
      fail("unexpected exception " + e);
    }
  }

  /**
   * Test suite.
   * @return test instance
   */
  public static
  Test suite() {
    return new TestSuite(BufferedRandomFileInputStreamTest.class);
  }
}


//======================================================================
//
// $Log: BufferedRandomFileInputStreamTest.java,v $
// Revision 1.1  1999/08/03 11:31:22  gunterma
// The version for initial checkin.
//
