/*
 * $Id: PP4Pages.java,v 1.6 2000/05/03 09:49:20 gunterma Exp $
 * (History at end)
 */

/*
 * This file is part of the PPower4 software
 * Copyright (c) 1999,2000  Klaus Guntermann/TU Darmstadt, FB 20, FG SP
 *
 * See the file COPYING in the top level directory for details.
 */

/**
   Process pdf files and trigger commands, when comments are found
   within a page.
   @author Klaus Guntermann
*/
package de.tu_darmstadt.sp.pp4;

import java.io.*;
import java.util.*;
import java.text.*;
import com.etymon.pj.*;
import com.etymon.pj.exception.*;
import com.etymon.pj.object.*;

import de.tu_darmstadt.sp.pp4.PP4Handler;


public class PP4Pages {
  private Pdf inpdf; // our pdf file to work with
  // Some of the internal structures are needed several times. Keep them here
  private PjCatalog catalog;
  private PjPage thepage=null; // the current page as it is processed...
  private int pageref=0;

  private String progid=null;
  // We need to keep the arguments of a page split operation.
  private String pageArguments=null;

  private String command=null, arguments=null;
  // the active/last comment processed.
  private int bgobj=0, nextpagebgobj=0;
  // a background object, if there is any for the current page and a
  // place to save it, if we cannot put it into the current page now.

  private StringBuffer streambuf=null;
      // the string to be written to partial pages

  private int numpages=0;
  private Vector destNames=null;
  private String curnum=null;
  private String inpdfname=null, // the input file name
    outpdfname=null; // the destination file name for the result
  private int verboseLevel=0, charsinline=0; // guess ...
  private int compress=1; // indication for output format
  private PrintStream os=null; // the stream for messages to be shown
  private Hashtable handlerhash = new Hashtable();
  private boolean doCopy=true, pageAlreadyWritten=false;
  private Vector freeStreamNumbers = new Vector();
  private Vector pageParts = new Vector();
  private Hashtable streamHash=new Hashtable();

  /** Create a new object, saving away the arguments.
      @param infile the name of the input pdf file.
      @param outfile the name of the output pdf file.
      @param outs the PrintStream for the message output.
      @param id the name and version to tag the output.
  */
  public PP4Pages(String infile, String outfile, PrintStream outs, String id)
    {
      // Save all arguments, which will be referenced soon
      inpdfname = infile;
      outpdfname = outfile;
      if (os==null) os = outs;
      progid = id;
    }


  /**
     Get the current copying state.
     @return the boolean value
  */
  public boolean getCopyingState()
    {
      return(doCopy);
    }


  /**
     Set the current copying state.
     @param state for next actions
  */
  public void setCopyingState(boolean state)
    {
      doCopy=state;
    }


  /**
     Set the current level of verbosity for tracing.
     @param level new verbosity
  */
  public void setVerboseLevel(int level)
    {
      verboseLevel=level;
    }


  /**
     Set the current level of compression.
     Currently only levels less than or equal to 0 (no compression)
     and greater 0 (default compression) are distinguished.
     @param level new compression
  */
  public void setCompressLevel(int level)
    {
      compress=level;
    }


  // Track the progress by writing page numbers and dots.
  // Arguments: 1. initial string
  //            2. optional page number (if > 0)
  private void traceprogress(String txt, int n)
    {
      if (n>0) { // write string and page number
	os.print(txt+n);
	charsinline+=1;
	int no=n; // how many places did the number take?
	while (no>0) {
	  charsinline+=1; no/=10;
	}
      } else { // only a string. Maybe we have to start a new line for
	// better readability
	if (charsinline>72) {
	  os.println(txt);
	  charsinline=0;
	} else {
	  os.print(txt);
	  charsinline+=1;
	}
      }
    }


  /**
     Add comment processing and give the corresponding handler.
     @param hd the handler object to activate, when the match is
        triggered.
  */
  public void addCommentProcessor(PP4Handler hd)
    {
      // Make sure, that the string is not yet in the table, then add
      // it eventually.
      String pattern=hd.pattern();
      hd.setPagesRef(this);
      if (verboseLevel>3) os.println("register pattern "+pattern);
      if (handlerhash.get(pattern)!=null) {
	os.println("Warning: duplicate pattern "+pattern+" not registered.");
	return;
      }
      handlerhash.put(pattern,hd);
    }


  /**
     Get rid of the contents of the current stream's copy.
  */
  public void clearStreamBuffer()
    {
      streambuf=new StringBuffer(8192);
    }


  /**
     Append a string to the stream buffer for the current stream.
     Must take care not to break a command currently in progress.
     @param contents the string to add.
  */
  public void append(String contents)
    {
      streambuf.append(contents);
    }


  /**
     Prepend a string to the current stream, will be seen in
     all copies of the page, which are not yet written.
     @param context the string to add.
  */
  public void prepend(String context)
    {
      streambuf.insert(0,context);
    }


  /**
     Create a stream in the current document with the given content.
     Take care: We try to reuse stream numbers collected in
     <code>freeStreamNumbers</code>. This will have the effect of
     replacing a stream in the file, if a stream was freed just before.
     @param txt the contents for the stream to create.
     @return an identification of the added part, such that it can be
       used later in <code>addBgStream</code>.
  */
  public Object makeStream(String txt)
    {
      Integer number;
      PjStream stream1 =new PjStream(txt.getBytes());
      if (compress>0) try {
	  stream1 = stream1.flateCompress();
      }
      catch (PjException pje) {
	os.println(pje); return (null);
      }
      // See whether we can reuse an existing stream...
      int numberOfFree=freeStreamNumbers.size();
      if (numberOfFree>0) {
	// Yes, we can. Determine the number from the Vector and use
	// the last number, to replace a just freed element first!
	number=(Integer)freeStreamNumbers.lastElement();
	freeStreamNumbers.removeElementAt(numberOfFree-1);
	inpdf.registerObject(stream1,number.intValue());
      } else // No, we must allocate a new stream number.
	number=new Integer(inpdf.registerObject(stream1));
      return(number);
    }


  /**
     Add a stream as a background to the current page, all subsequent
     pages and all copies of it. Does not work immediately, if a copy
     of the current page has been written before.
     @param obj determines the stream to add. The argument should be
       the result of <code>makeStream</code>
  */
  public void addBgStream(Object obj)
    {
      if (obj instanceof Integer)
	if (pageAlreadyWritten) {
	  // We should not put it into the current page now, because
	  // there are already parts of the page out.
	  nextpagebgobj=((Integer)obj).intValue();
	} else {
	  bgobj=((Integer)obj).intValue();
	  // This will be the background for all subsequent pages.
	  // We cannot do anything about this now.
	}
      // otherwise we have a bad usage...
    }


  /**
     Replace the current stream contents by the contents of the page buffer.
  */
  public void addStream()
    {

      Integer newstream=(Integer)makeStream(streambuf.toString());
      // And add it to the vector of parts of this page.
      pageParts.addElement(newstream);
      clearStreamBuffer(); // contents are saved in a stream now.

    }


  // get the destination entries for names or return null, if
  // this is not available.
  private Vector getDestNames(Pdf inpdf)
    {
      PjCatalog c;
      Hashtable ht;
      PjObject o;
      //os.println("getDestNames");
      if (inpdf==null) return null;
      try {
	c =  (PjCatalog) inpdf.getObject(inpdf.getCatalog());
	if (c==null) return null;
	//os.println("got Root");
	o = c.getNames();
	if (o==null) return null;
	//os.println("got names");
	if (o instanceof PjReference) o=inpdf.resolve(o);
	if (o instanceof PjDictionary) ht=((PjDictionary)o).getHashtable();
	else return null;
	//os.println("try /Dests");
	o = (PjObject)ht.get(PjName.DESTS);
	if (o==null) return null;
	//os.println("got dests");
	if (o instanceof PjReference) o=inpdf.resolve(o);
	if (o instanceof PjDictionary) ht=((PjDictionary)o).getHashtable();
	else return null;
	//os.println("try kids");
	o = (PjObject)ht.get(PjName.KIDS);
	if (o==null) return null;
	//os.println("got kids");
	if (o instanceof PjReference) o=inpdf.resolve(o);
	if (o instanceof PjArray) {
	  //os.println("try elem of kids");
	  o =(PjObject)(((PjArray)o).getVector().elementAt(0));
	  if (o==null) return null;
	  //os.println("got elem of kids");
	  if (o instanceof PjReference) o=inpdf.resolve(o);
	  if (o instanceof PjDictionary) ht=((PjDictionary)o).getHashtable();
	  else return null; 
	  o = (PjObject)ht.get(PjName.NAMES);
	  if (o==null) return null;
	  //os.println("got names");
	  if (o instanceof PjReference) o=inpdf.resolve(o);
	  if (o instanceof PjArray) {
	    //os.println("done.:"+o);
	    return ((PjArray)o).getVector();
	  } //else os.println("names not an array");
	} //else  os.println("kids not an array");
      } catch (PjException pje) {
	os.println(pje);
      }
      return null;
    }

  // get the entry from the pages dictionary
  // We must make sure to use only the page entries. Otherwise the
  // numbering may be rather strange.
  private String getPageDestNum(PjPage page)
    {
      String val=null;
      // do we have the information?
      if (destNames==null) return null;
      if (page==null) return null;
      // scan the Dictionary to find the entry for page
      int numelem=destNames.size();
      PjName dName=new PjName("D");
      for (int inx=1; inx<numelem; inx+=2) {
	Object o=destNames.elementAt(inx);
	if (o instanceof PjReference) o=inpdf.resolve((PjReference)o);
	if (o!=null && o instanceof PjDictionary) {
	  // os.print("scanning,");
	  o=(PjObject)(((PjDictionary)o).getHashtable().get(dName));
	  if (o instanceof PjReference) o=inpdf.resolve((PjReference)o);
	  if (o!=null && o instanceof PjArray) {
	    PjObject p=(PjObject)(((PjArray)o).getVector().elementAt(0));
	    if (((PjPage)inpdf.resolve(p))==page) {
	      o=destNames.elementAt(inx-1);
	      if (o instanceof PjString) {
		String s=((PjString)o).getString();
		//os.println("check "+s);
		if (s.startsWith("page.")) {
		  //os.println("matched.");
		  return s;
		}
	      } //else os.println("no PjString value, but "+o+".");
	    } //else os.println("no match");
	  } //else os.println("no Array or null: "+o+".");
	}
      }
      //os.println("No match (in "+numelem+" elements).");
      return null;
    }

  // add a new entry to the dictionary
  // CHECK: multiple entries?
  private void putPageDestNum(int pageref, String number)
    {
      //os.print("putPageDestNum:");
      // do nothing, if we do not have the names array
      if (destNames==null) return;
      // we must create a new Dictionary describing the location on
      // the page. As we do not know any special place, lets move to
      // location (0,0)
      PjDictionary dict=new PjDictionary();
      Vector v=new Vector(5);
      dict.getHashtable().put(new PjName("D"),new PjArray(v));
      v.addElement(new PjReference(new PjNumber(pageref)));
      v.addElement(new PjName("XYZ"));
      v.addElement(new PjNumber(0));
      v.addElement(new PjNumber(0));
      v.addElement(new PjNull());
      int dictnum=inpdf.registerObject(dict);
      PjReference dictref=new PjReference(new PjNumber(dictnum));
      // replace this, if we have this entry already
      int numelem=destNames.size();
      for (int inx=0; inx<numelem; inx+=2) {
	PjObject o=(PjObject)destNames.elementAt(inx);
	//os.println("string scanning at "+o+".");
	if (o instanceof PjString) {
	  String s=((PjString)o).getString();
	  if (s.equals(number)) {
	    //os.println("entry exists, replace it.");
	    destNames.setElementAt(dictref,inx+1);
	    return; // done...
	  } //else os.println("failed "+s+"/"+number+".");
	}
      }
      destNames.addElement(new PjString(number));
      destNames.addElement(dictref);
    }


  /**
     Make the current list of page chunks the contents of the given
     page.
     @param thisPage the page to fill.
  */
  private void chunksToPage(PjPage thisPage)
    {
      try {
	// We do not need to create a new stream object, but we have to
	// replace the contents.

	// Make the page contents an empty array.
	thisPage.setContents(new PjArray());

	// Then insert the parts for this page.
	if (bgobj!=0)
	  inpdf.addToPage(thisPage,bgobj);
	for (int num=0; num<pageParts.size();num++)
	  inpdf.addToPage(thisPage,((Integer)pageParts.elementAt(num)).
			  intValue());

      }
      catch (PjException pje) {
	os.println(pje);
      }
    }


  /**
     Write a new intermediate page from the page buffers to
     the pdf file.
  */
  public void makeNewPage(String arguments)
    {
      try {
	// Create a new page.
	PjPage page = new PjPage();

	// This page needs the same Resources, Mediabox, Annots and
	// Trans as the master page. (any other things omitted?)
	page.setResources((PjReference)thepage.getResources());
	page.setMediaBox((PjRectangle)thepage.getMediaBox());
	PjArray annots =(PjArray)thepage.getAnnots();
	if (annots!=null) {// set annots only, if present.
	  // Annots are present, e.g. if the page has active hyperlinks.
	  page.setAnnots(annots);
	}
	if (pageArguments!=null) {
	  setTransition(page,pageArguments);
	} else {
	  PjObject trans =thepage.getTrans();
	  if (trans!=null) {// set trans only, if present.
	    // trans is present, e.g. if the page has a special effect when
	    // we go to it
	    if (trans instanceof PjDictionary)
	      page.setTrans((PjDictionary)trans);
	    else if (trans instanceof PjReference)
	      page.setTrans((PjReference)trans);
	    // otherwise ignore this trans entry. What is it???
	  }
	}

	// We need to build a new stream, if we have some contents to
	// display in our buffer.
	if (streambuf.length()>0)
	  // Create the contents for this...
	  addStream();

	// Write the collected streams for the new page.
	chunksToPage(page);
	int newpref; // we need a page reference...
	inpdf.appendPage(newpref=inpdf.registerObject(page));

	// If this was the first chunk, we need to add a target
	// for referring to this page.
	if (!pageAlreadyWritten && curnum!=null)
	  putPageDestNum(newpref, "first"+curnum);

	// any arguments for the next page?
	if (arguments!=null)
	  if (arguments.length()>0) {
	    arguments=arguments.trim();
	    if (arguments.length()>7 &&
		arguments.startsWith("trans='")&&arguments.endsWith("'")) {
	      pageArguments=arguments.substring(7,arguments.length()-1);
	    } else {
	      os.println("Warning: argument "+arguments+" not processed.");
	    }
	  }

	if (verboseLevel>0) {
	  traceprogress(".",0);
	}
	pageAlreadyWritten=true;
      }
      catch (PjException pje) {
	os.println(pje);
      }
    }

  /**
     Replace or add a transition entry to the given page.
     We have to decode the arguments.
  */
  private void setTransition(PjPage page, String pageArguments)
    {
      PjDictionary transition=new PjDictionary();
      int parStartPos=0,
	  parEndPos=pageArguments.indexOf(','),
	  index=-1, defIndex;
      String subType;
      String [] param;
      param = new String[3];
      // first get the type of the transition and the arguments
      if (parEndPos<0) { // Just the command
	subType=pageArguments;
      } else {// command and params
	subType=pageArguments.substring(0,parEndPos);
	parStartPos=parEndPos+1;
	for (index=-1; parStartPos>0; parStartPos=parEndPos+1) {
	  parEndPos=pageArguments.indexOf(',',parStartPos);
	  if (index<2) {
	    if (parEndPos<0) {
	      param[++index]=pageArguments.substring(parStartPos);
	    } else {
	      param[++index]=pageArguments.substring(parStartPos,parEndPos);
	    }
	  } // otherwise ignore values
	}
      }
      if (verboseLevel>5) {
	  os.print("processing command: '"+subType);
	  for (int i=0;i<=index;i++) os.print("/"+param[i]);
	  os.println("'"+index+".");
      }
      if (subType!=null && subType.length()>0)
	transition.getHashtable().put(new PjName("S"), new PjName(subType));
      // normally we have no parameters
      defIndex=-1;
      // if it has parameters, process them
      if (subType.compareTo("Split")==0) {
	defIndex=1;
	if ( index<1 || param[0].length()<1 || param[1].length()<1 ) {
	  os.println("Warning: missing argument: "+pageArguments+".");
	  return;
	}
	transition.getHashtable().put(new PjName("Dm"), new PjName(param[0]));
	transition.getHashtable().put(new PjName("M"), new PjName(param[1]));
      } else
      if (subType.compareTo("Box")==0) {
	defIndex=0;
	if (index<0 || param[0].length()<1) {
	  os.println("Warning: missing argument: "+pageArguments+".");
	  return;
	}
	transition.getHashtable().put(new PjName("M"), new PjName(param[0]));
      } else
      if (subType.compareTo("Blinds")==0) {
	defIndex=0;
	if (index<0 || param[0].length()<1) {
	  os.println("Warning: missing argument: "+pageArguments+".");
	  return;
	}
	transition.getHashtable().put(new PjName("Dm"), new PjName(param[0]));
      } else
      if (subType.compareTo("Glitter")==0 || subType.compareTo("Wipe")==0) {
	defIndex=0;
	if (index<0 || param[0].length()<1) {
	  os.println("Warning: missing argument: "+pageArguments+".");
	  return;
	}
	try {
	  transition.getHashtable().put(new PjName("Di"),
			new PjNumber(Float.valueOf(param[0]).floatValue()));
	} catch (NumberFormatException nfe) {
	  System.out.println("Warning: expected number, got "+param[0]);
	}
      } else if (subType.compareTo("Dissolve")!=0 &&
		 subType.compareTo("Replace")!=0) {
	  os.println("Warning: Unknown transition effect: "+pageArguments+".");
	  return;
      }
      // Maybe there is additionally an optional duration specified
      if (index>defIndex) {
	try {
	  transition.getHashtable().put(new PjName("D"),
		new PjNumber(Float.valueOf(param[defIndex+1]).floatValue()));
	} catch (NumberFormatException nfe) {
	  System.out.println("Warning: expected number, "+param[index]);
	}
	if (index>defIndex+1) {
	  os.println("Warning: too many arguments for "+subType);
	}
      }
      // now update the transition info
      if (verboseLevel>5) {
	  os.print("transition:"+transition+".");
      }
      page.setTrans(transition);
    }

  // this is an internal routine to be called for all parts of a page
  private void scanPageStream(PjReference theStream)
    {
      Vector currentChunks=(Vector)streamHash.get(theStream);
      if (currentChunks==null) {
	// This stream has not yet been processed.
	int oldchunkNumber=pageParts.size(); // keep track of new chunks
	PjStream pageStream=(PjStream)inpdf.resolve(theStream);
	// Get the contents of the page, maybe uncompress it and make it
	// a stream to read line by line, because the comments which we
	// are going to parse, start on a line by themselves...
	try {
	  BufferedReader readstream = new BufferedReader(new
		 InputStreamReader(new ByteArrayInputStream(
		    ((PjStream)pageStream.flateDecompress()).getBuffer())));
	  // Run through the page, searching for the comments to process.
	  String current;
	  while ((current=readstream.readLine())!=null) {
	    if (current.length()==0) continue; // ignore empty lines
	    if (current.charAt(0)=='%') {// this is a comment to process
	      if (verboseLevel>4)
		os.println("Found comment:["+current+"]");
	      int spacepos=current.indexOf(' ');
	      if (spacepos==-1) {
		spacepos=current.length();
		arguments = "";
	      } else arguments=current.substring(spacepos+1,current.length());
	      command=current.substring(0,spacepos);
	      PP4Handler hd = (PP4Handler) handlerhash.get(command);
	      if (hd!=null) {
		try {
		  hd.doit(command,arguments);
		} catch (PP4Exception e) {
		  // Show that we got an exception.
		  os.println(e);
		}
	      } else if (doCopy) append(current+"\n");
	    } else if (doCopy) append(current+"\n");
	  } // while
	  // Now we do not need this element any more and can add its
	  // pdf object number into our pool of free streams, whose
	  // numbers can be reused. This will remove the old stream
	  // contents from our output file, too.
	  freeStreamNumbers.addElement(new Integer(theStream.
						   getObjNumber().getInt()));
	  // Make that final stream part complete.
	  // (Replace the current stream by the new contents.)
	  addStream();
	  // Note the chunks from oldchunksNumber to pageParts.size()
	  if (oldchunkNumber==0) {
	    // the easy part, only chunks from our stream on this page
	    // take care: must clone the vector, to keep this state
	    streamHash.put(theStream,pageParts.clone());
	  } else {//need a new Vector with the chunks from this stream
	    Vector theseChunks=new Vector(pageParts.size()-oldchunkNumber);
	    for (int i=oldchunkNumber;i<pageParts.size();i++) {
	      theseChunks.addElement(pageParts.elementAt(i));
	    }
	    // take care: must clone the vector, to keep this state
	    streamHash.put(theStream,theseChunks.clone());
	  }
	}
	catch (PjException pje) {
	  os.println(pje);
	}
	catch (IOException ioe) {
	  System.out.println(ioe);
	}
      } else {
	// Chunks were already defined, and possibly we got more than
	// one.
	// Insert them one by one into the list of this page.
	pageParts.addElement(currentChunks.elementAt(0));
	for (int i=1;i<currentChunks.size();i++) {
	  makeNewPage(null);
	  pageParts.addElement(currentChunks.elementAt(i));
	}
      }
    }


  /**
    Deliver the bounding box parameters of the current page.
    @return array of four <code>float</code> values "LLX LLY URX URY"
  */
  public float [] getPageBounds()
    {

      PjRectangle mbox=null;
      try {
	mbox=(PjRectangle)thepage.getMediaBox();
      }
      catch (PjException pje) {
	os.println(pje); return (null);
      }
      float [] bounds = new float[4];

      bounds[0]=mbox.getLowerLeftX().getFloat();
      bounds[1]=mbox.getLowerLeftY().getFloat();
      bounds[2]=mbox.getUpperRightX().getFloat();
      bounds[3]=mbox.getUpperRightY().getFloat();
      return(bounds);
    }


  /**
     Scan through the pages of the document and then write the result.
  */
  public void work()
    {
      try {
	// Initialize access to the pdf file.
	if (verboseLevel>1) {
	  os.println("Reading and parsing input file...");
	}
	// Load the input file. (This may take some time for large
	// files; please be patient.)
	inpdf = new Pdf(inpdfname);
	catalog = (PjCatalog) inpdf.getObject(inpdf.getCatalog());

	destNames=getDestNames(inpdf);

	numpages = inpdf.getPageCount();

	if (verboseLevel>1) {
	  os.println("Got it ("+numpages+" Pages).");
	}
	for (int pageno=1;pageno<=numpages;pageno++) {
	  // We fetch always the first page because the original will be
	  // deleted after processing. This makes the next page the first.
	  pageref = inpdf.getPage(1);
	  thepage = (PjPage)inpdf.getObject(pageref);

	  if (verboseLevel>0) {
	    traceprogress("[",pageno);
	  }

	  pageAlreadyWritten=false;

	  // find the internal number for this page
	  curnum=getPageDestNum(thepage);
	  if (curnum!=null) {
	    int i=curnum.indexOf('.');
	    if (i>0) curnum=curnum.substring(i);
	  }

	  clearStreamBuffer();
	  pageParts.removeAllElements();
	  if (nextpagebgobj!=0) {
	    bgobj=nextpagebgobj; nextpagebgobj=0;
	  }
	  PjObject contents = thepage.getContents();

	  PP4Handler hd = (PP4Handler) handlerhash.get("BOPHOOK");
	  if (hd!=null) try {
	    hd.doit("BOPHOOK",""+pageno);
	  } catch (PP4Exception e) {
	    // Show that we got an exception.
	    os.println(e);
	  }
	  // must check whether we have a single reference or an array
	  // of references (does pdftex create that?)
	  if (contents instanceof PjReference) {
	    scanPageStream((PjReference)contents);
	  } else if (contents instanceof PjArray) {
	    // If we have an array, go through all elements
	    Vector vector=((PjArray)contents).getVector();
	    for (int scount=0;scount<vector.size();scount++) {
	      scanPageStream((PjReference)vector.elementAt(scount));
	    }
	  } else  {
	    os.println("Unexpected type of contents object for page "+
		       pageno+".");
	  }

	  hd = (PP4Handler) handlerhash.get("EOPHOOK");
	  if (hd!=null) try {
	    hd.doit("EOPHOOK",""+pageno);
	  } catch (PP4Exception e) {
	    // Show that we got an exception.
	    os.println(e);
	  }

	  // Show all parts of this page again.
	  chunksToPage(thepage);

	  // Let's see, whether we need to modify transition and reset
	  // the triggering saved value. We do not want to influence
	  // the real next page.
	  if (pageArguments!=null) {
	    setTransition(thepage,pageArguments);
	    pageArguments=null;
	  }
	  // And append the page again to the list of pages.
	  inpdf.appendPage(pageref);
	  // if we did not create copies, we need to insert the link
	  if (!pageAlreadyWritten && curnum!=null) {
	    putPageDestNum(pageref, "first"+curnum);
	  }
	  // Delete the original copy.
	  inpdf.deletePage(1);

	  if (verboseLevel>0) {// Show that this page is done.
	    traceprogress("]",0);
	  }
	} // for

	// we have to do some updates, e.g. replace the OpenAction in
	// the Catalog. It should point to the new first page. This is
	// currently not supported in the library. Thus access the
	// information manually.
	PjObject action = catalog.getOpenAction();
	if (action instanceof PjReference) action=inpdf.resolve(action);
	if (action instanceof PjDictionary) {
	  Hashtable openactionhash;
	  openactionhash=((PjDictionary)action).getHashtable();
	  PjName name =new PjName("D");
	  if (openactionhash.containsKey(name)) {
	    PjArray startpage = (PjArray) openactionhash.get(name);
	    int pageref = inpdf.getPage(1);
	    if (verboseLevel>4) {
	      os.print("Replacing startpage in OpenAction /D by "+
			       pageref+" 0 R.");
	    }
	    startpage.getVector().setElementAt(
			       new PjReference(new PjNumber(pageref)),0);
	  }
	} else if (action!=null) {
	  PjArray startpage = (PjArray) action;
	  int pageref = inpdf.getPage(1);
	  if (verboseLevel>4) {
	    os.print("Replacing startpage in OpenAction array by "+
			     pageref+" 0 R.");
	  }
	  startpage.getVector().setElementAt(
			     new PjReference(new PjNumber(pageref)),0);
	}

	// We should also leave some traces in the Info...
	PjInfo info=(PjInfo)inpdf.resolve(inpdf.getInfoDictionary());
	PjObject creator=info.getCreator();
	if (creator instanceof PjReference) {
	  creator=inpdf.resolve((PjReference)creator);
	}
	info.setCreator(new PjString(((PjString)creator).getString()+
			" extended by "+progid));

	// Write the resulting file.
	if (verboseLevel>0) {
	  os.println("\nWriting modified output file...");
	}
	// Do that with buffered stream to speed up.
	BufferedOutputStream outstream =
	  new BufferedOutputStream(new FileOutputStream(outpdfname),8192);
	inpdf.writeToStream(outstream);
	outstream.close();
	if (verboseLevel>0) {
	  os.println("Done.");
	}
      }
      catch (PjException pje) {
	System.out.println(pje);
      }
      catch (IOException ioe) {
	System.out.println(ioe);
      }
    }
}

/*
 * $Log: PP4Pages.java,v $
 * Revision 1.6  2000/05/03 09:49:20  gunterma
 * fix problem with empty lines in pdf stream objects
 *
 * Revision 1.5  2000/04/12 06:09:03  gunterma
 * Fix a problem with inserting the "first" markers.
 *
 * Revision 1.4  2000/04/10 06:35:39  gunterma
 * make this version compatible with Java 1.1.x
 *
 * Revision 1.3  2000/04/03 14:13:05  gunterma
 * Optional arguments processing fixed.
 * Added --version/-V flag to main program.
 * PPower4 version 0.8 to be released.
 *
 * Revision 1.2  2000/03/13 09:48:14  gunterma
 * Allow optional arguments when processing a pause to define a new
 * transition mode for the created page.
 *
 * Revision 1.1  1999/08/09 11:56:48  gunterma
 * initial checkin
 *
 */
